@@ -1,6 +1,6 @@
 /* 
 **  mod_layout.c -- Apache layout module
-**  $Revision: 1.129 $
+**  $Revision: 1.141 $
 */
 
 #include "mod_layout.h"
@@ -13,7 +13,9 @@ static void *create_dir_mconfig(pool *p, char *dir) {
 	on. 
 */
 	layout_conf *cfg;
+
 	cfg = ap_pcalloc(p, sizeof(layout_conf));
+	cfg->dir = ap_pstrdup(p, dir);
 	cfg->replace_tags = UNSET;
 	cfg->merge = UNSET;
 	cfg->notes = UNSET;
@@ -25,10 +27,9 @@ static void *create_dir_mconfig(pool *p, char *dir) {
 	cfg->display_origin = ON;
 	cfg->header_enabled = UNSET;
 	cfg->footer_enabled = UNSET;
+	cfg->pattern_enabled = UNSET;
 	cfg->http_header_enabled = UNSET;
 	cfg->http_header = NULL;
-	cfg->header = NULL;
-	cfg->footer = NULL;
 	cfg->async_cache = LAYOUT_CACHE;
 	cfg->begin_tag = LAYOUT_BEGINTAG;
 	cfg->end_tag = LAYOUT_ENDTAG;
@@ -39,9 +40,10 @@ static void *create_dir_mconfig(pool *p, char *dir) {
 	cfg->uris_ignore_header = NULL;
 	cfg->uris_ignore_http_header = NULL;
 	cfg->uris_ignore_footer = NULL;
-	cfg->merge_ignore = NULL;
-	cfg->merge_ignore_footer = NULL;
-	cfg->merge_ignore_header = NULL;
+	cfg->tag_ignore = NULL;
+	cfg->tag_ignore_footer = NULL;
+	cfg->tag_ignore_header = NULL;
+	cfg->layouts = NULL;
 	cfg->layout_html_handler = ON;
 
 	ap_table_setn(cfg->types, INCLUDES_MAGIC_TYPE, "1");
@@ -55,7 +57,6 @@ static void *create_dir_mconfig(pool *p, char *dir) {
 	ap_table_setn(cfg->types, "application/x-httpd-php", "1");
 	ap_table_setn(cfg->types, "LayoutHandler application/x-httpd-php3", "1");
 	ap_table_setn(cfg->types, "LayoutHandler application/x-httpd-php3-source", "1");
-
 	return (void *) cfg;
 }
 
@@ -63,17 +64,19 @@ static void *merge_dir_mconfig(pool *p, void *origin, void *new) {
 	layout_conf *cfg = ap_pcalloc(p, sizeof(layout_conf));
 	layout_conf *cfg_origin = (layout_conf *)origin;
 	layout_conf *cfg_new = (layout_conf *)new;
+	cfg->dir = ap_pstrdup(p, cfg_new->dir);
 	cfg->replace_tags = UNSET;
 	cfg->merge = UNSET;
 	cfg->notes = UNSET;
 	cfg->proxy = UNSET;
 	cfg->comment = UNSET;
+	cfg->append_header = UNSET;
+	cfg->append_footer = UNSET;
 	cfg->cache_needed = UNSET;
-	cfg->header = NULL;
-	cfg->footer = NULL;
 	cfg->display_origin = ON;
 	cfg->header_enabled = UNSET;
 	cfg->footer_enabled = UNSET;
+	cfg->pattern_enabled = UNSET;
 	cfg->http_header_enabled = UNSET;
 	cfg->http_header = NULL;
 	cfg->time_format = LAYOUT_TIMEFORMAT;
@@ -81,32 +84,34 @@ static void *merge_dir_mconfig(pool *p, void *origin, void *new) {
 	cfg->async_cache = LAYOUT_CACHE;
 	cfg->begin_tag = LAYOUT_BEGINTAG;
 	cfg->end_tag = LAYOUT_ENDTAG;
-	cfg->append_header = UNSET;
-	cfg->append_footer = UNSET;
-	cfg->merge_ignore = NULL;
-	cfg->merge_ignore_footer = NULL;
-	cfg->merge_ignore_header = NULL;
+	cfg->uris_ignore_header = NULL;
+	cfg->uris_ignore_http_header = NULL;
+	cfg->uris_ignore_footer = NULL;
+	cfg->tag_ignore = NULL;
+	cfg->tag_ignore_footer = NULL;
+	cfg->tag_ignore_header = NULL;
+	cfg->layouts = NULL;
 	cfg->layout_html_handler = ON;
-	
-	if(strcmp(cfg_new->async_cache, LAYOUT_CACHE)){
+
+	if (strcmp(cfg_new->async_cache, LAYOUT_CACHE)){
 		cfg->async_cache = ap_pstrdup(p, cfg_new->async_cache);
 	} else if (strcmp(cfg_origin->async_cache, LAYOUT_CACHE)){
 		cfg->async_cache = ap_pstrdup(p, cfg_origin->async_cache);
 	}
 
-	if(strcmp(cfg_new->time_format, LAYOUT_TIMEFORMAT)){
+	if (strcmp(cfg_new->time_format, LAYOUT_TIMEFORMAT)){
 		cfg->time_format = ap_pstrdup(p, cfg_new->time_format);
 	} else if (strcmp(cfg_origin->time_format, LAYOUT_TIMEFORMAT)){
 		cfg->time_format = ap_pstrdup(p, cfg_origin->time_format);
 	}
 
-	if(strcmp(cfg_new->begin_tag, LAYOUT_BEGINTAG)){
+	if (strcmp(cfg_new->begin_tag, LAYOUT_BEGINTAG)){
 		cfg->begin_tag = ap_pstrdup(p, cfg_new->begin_tag);
 	} else if (strcmp(cfg_origin->begin_tag, LAYOUT_BEGINTAG)){
 		cfg->begin_tag = ap_pstrdup(p, cfg_origin->begin_tag);
 	}
 
-	if(strcmp(cfg_new->end_tag, LAYOUT_ENDTAG)){
+	if (strcmp(cfg_new->end_tag, LAYOUT_ENDTAG)){
 		cfg->end_tag = ap_pstrdup(p, cfg_new->end_tag);
 	} else if (strcmp(cfg_origin->end_tag, LAYOUT_ENDTAG)){
 		cfg->end_tag = ap_pstrdup(p, cfg_origin->end_tag);
@@ -125,41 +130,36 @@ static void *merge_dir_mconfig(pool *p, void *origin, void *new) {
 	cfg->append_header = (cfg_new->append_header == UNSET) ? cfg_origin->append_header : cfg_new->append_header;
 	cfg->append_footer = (cfg_new->append_footer == UNSET) ? cfg_origin->append_footer : cfg_new->append_footer;
 
-	if(isOn(cfg->append_header))  {
-		if(isOn(cfg_new->header_enabled) || isOn(cfg_origin->header_enabled)) {
-			cfg->header_enabled = ON;
-			cfg->header = ap_append_arrays(p, cfg_origin->header, cfg_new->header);
-		}
-	} else {
-		if(cfg_new->header_enabled == UNSET){
-			cfg->header = cfg_origin->header;
-			cfg->header_enabled = cfg_origin->header_enabled;
-		} else if (isOn(cfg_new->header_enabled)){
-			cfg->header = cfg_new->header;
-			cfg->header_enabled = cfg_new->header_enabled;
-		} else {
-			cfg->header_enabled = OFF;
-		}
-	}
-
-	if(isOn(cfg->append_footer))  {
-		if(isOn(cfg_new->footer_enabled) || isOn(cfg_origin->footer_enabled)) {
-			cfg->footer_enabled = ON;
-			cfg->footer = ap_append_arrays(p, cfg_origin->footer, cfg_new->footer);
-		}
-	} else {
-		if(cfg_new->footer_enabled == UNSET){
-			cfg->footer = cfg_origin->footer;
-			cfg->footer_enabled = cfg_origin->footer_enabled;
-		} else if (isOn(cfg_new->footer_enabled)){
-			cfg->footer = cfg_new->footer;
+	if (isOn(cfg->append_header) || isOn(cfg->append_footer))  {
+		if (isOn(cfg->append_header) && isOn(cfg->append_footer))  {
+			cfg->layouts = ap_append_arrays(p, cfg_origin->layouts, cfg_new->layouts);
+			cfg->header_enabled = cfg_new->header_enabled ? cfg_new->header_enabled : cfg_origin->header_enabled;
+			cfg->footer_enabled = cfg_new->footer_enabled ? cfg_new->footer_enabled : cfg_origin->header_enabled;
+		} else if (isOn(cfg->append_header)) {
+			cfg->header_enabled = cfg_new->header_enabled ? cfg_new->header_enabled : cfg_origin->header_enabled;
 			cfg->footer_enabled = cfg_new->footer_enabled;
+			cfg->layouts = layout_array_push_kind(p, cfg_origin->layouts, cfg_new->layouts, HEADER);
+		/* We now assume just append_footer */
+		} else  {
+			cfg->header_enabled = cfg_new->header_enabled;
+			cfg->footer_enabled = cfg_new->footer_enabled ? cfg_new->footer_enabled : cfg_origin->header_enabled;
+			cfg->layouts = layout_array_push_kind(p, cfg_origin->layouts, cfg_new->layouts, FOOTER);
+		}
+	} else {
+		if (cfg_new->layouts){
+			cfg->layouts = cfg_new->layouts;
+			cfg->header_enabled = cfg_new->header_enabled;
+			cfg->footer_enabled = cfg_new->footer_enabled;
+			cfg->pattern_enabled = cfg_new->pattern_enabled;
 		} else {
-			cfg->footer_enabled = OFF;
+			cfg->layouts = cfg_origin->layouts;
+			cfg->header_enabled = cfg_origin->header_enabled;
+			cfg->footer_enabled = cfg_origin->footer_enabled;
+			cfg->pattern_enabled = cfg_origin->pattern_enabled;
 		}
 	}
 
-	if(cfg_new->http_header_enabled == UNSET){
+	if (cfg_new->http_header_enabled == UNSET){
 		cfg->http_header = ap_pstrdup(p, cfg_origin->http_header);
 		cfg->http_header_enabled = cfg_origin->http_header_enabled;
 	} else if (isOn(cfg_new->http_header_enabled)){
@@ -172,8 +172,8 @@ static void *merge_dir_mconfig(pool *p, void *origin, void *new) {
 	/* This is pretty simple */
 	cfg->types = cfg_new->types;
 
-	if(cfg_origin->uris_ignore) {
-		if(cfg_new->uris_ignore) {
+	if (cfg_origin->uris_ignore) {
+		if (cfg_new->uris_ignore) {
 			cfg->uris_ignore = ap_overlay_tables(p, cfg_new->uris_ignore, cfg_origin->uris_ignore);
 		} else {
 			cfg->uris_ignore = cfg_origin->uris_ignore;
@@ -182,8 +182,8 @@ static void *merge_dir_mconfig(pool *p, void *origin, void *new) {
 		cfg->uris_ignore = cfg_new->uris_ignore;
 	}
 
-	if(cfg_origin->uris_ignore_header) {
-		if(cfg_new->uris_ignore_header) {
+	if (cfg_origin->uris_ignore_header) {
+		if (cfg_new->uris_ignore_header) {
 			cfg->uris_ignore_header = ap_overlay_tables(p, cfg_new->uris_ignore_header, cfg_origin->uris_ignore_header);
 		} else {
 			cfg->uris_ignore_header = cfg_origin->uris_ignore_header;
@@ -192,8 +192,8 @@ static void *merge_dir_mconfig(pool *p, void *origin, void *new) {
 		cfg->uris_ignore_header = cfg_new->uris_ignore_header;
 	}
 
-	if(cfg_origin->uris_ignore_http_header) {
-		if(cfg_new->uris_ignore_http_header) {
+	if (cfg_origin->uris_ignore_http_header) {
+		if (cfg_new->uris_ignore_http_header) {
 			cfg->uris_ignore_http_header = ap_overlay_tables(p, cfg_new->uris_ignore_http_header, cfg_origin->uris_ignore_http_header);
 		} else {
 			cfg->uris_ignore_http_header = cfg_origin->uris_ignore_http_header;
@@ -202,8 +202,8 @@ static void *merge_dir_mconfig(pool *p, void *origin, void *new) {
 		cfg->uris_ignore_http_header = cfg_new->uris_ignore_http_header;
 	}
 
-	if(cfg_origin->uris_ignore_footer) {
-		if(cfg_new->uris_ignore_footer) {
+	if (cfg_origin->uris_ignore_footer) {
+		if (cfg_new->uris_ignore_footer) {
 			cfg->uris_ignore_footer = ap_overlay_tables(p, cfg_new->uris_ignore_footer, cfg_origin->uris_ignore_footer);
 		} else {
 			cfg->uris_ignore_footer = cfg_origin->uris_ignore_footer;
@@ -212,41 +212,41 @@ static void *merge_dir_mconfig(pool *p, void *origin, void *new) {
 		cfg->uris_ignore_footer = cfg_new->uris_ignore_footer;
 	}
 
-	if(cfg_origin->merge_ignore) {
-		if(cfg_new->merge_ignore) {
-			cfg->merge_ignore = ap_overlay_tables(p, cfg_new->merge_ignore, cfg_origin->merge_ignore);
+	if (cfg_origin->tag_ignore) {
+		if (cfg_new->tag_ignore) {
+			cfg->tag_ignore = ap_overlay_tables(p, cfg_new->tag_ignore, cfg_origin->tag_ignore);
 		} else {
-			cfg->merge_ignore = cfg_origin->merge_ignore;
+			cfg->tag_ignore = cfg_origin->tag_ignore;
 		}
 	} else {
-		cfg->merge_ignore = cfg_new->merge_ignore;
+		cfg->tag_ignore = cfg_new->tag_ignore;
 	}
 
-	if(cfg_origin->merge_ignore_footer) {
-		if(cfg_new->merge_ignore_footer) {
-			cfg->merge_ignore_footer = ap_overlay_tables(p, cfg_new->merge_ignore_footer, cfg_origin->merge_ignore_footer);
+	if (cfg_origin->tag_ignore_footer) {
+		if (cfg_new->tag_ignore_footer) {
+			cfg->tag_ignore_footer = ap_overlay_tables(p, cfg_new->tag_ignore_footer, cfg_origin->tag_ignore_footer);
 		} else {
-			cfg->merge_ignore_footer = cfg_origin->merge_ignore_footer;
+			cfg->tag_ignore_footer = cfg_origin->tag_ignore_footer;
 		}
 	} else {
-		cfg->merge_ignore_footer = cfg_new->merge_ignore_footer;
+		cfg->tag_ignore_footer = cfg_new->tag_ignore_footer;
 	}
 
-	if(cfg_origin->merge_ignore_header) {
-		if(cfg_new->merge_ignore_header) {
-			cfg->merge_ignore_header = ap_overlay_tables(p, cfg_new->merge_ignore_header, cfg_origin->merge_ignore_header);
+	if (cfg_origin->tag_ignore_header) {
+		if (cfg_new->tag_ignore_header) {
+			cfg->tag_ignore_header = ap_overlay_tables(p, cfg_new->tag_ignore_header, cfg_origin->tag_ignore_header);
 		} else {
-			cfg->merge_ignore_header = cfg_origin->merge_ignore_header;
+			cfg->tag_ignore_header = cfg_origin->tag_ignore_header;
 		}
 	} else {
-		cfg->merge_ignore_header = cfg_new->merge_ignore_header;
+		cfg->tag_ignore_header = cfg_new->tag_ignore_header;
 	}
 
 	return (void *) cfg;
 }
 
 static void layout_child_init(server_rec *s, pool *p) {
-	if(s->keep_alive) {
+	if (s->keep_alive) {
 		s->keep_alive = 0;
 	}
 
@@ -255,12 +255,10 @@ static void layout_child_init(server_rec *s, pool *p) {
 static int layout_handler (request_rec * r) {
 	int status;
 	int pid = 0;
-	int x = 0; /* used for increments */
 	layout_conf *cfg;
 	char *filename_post = NULL;
 	char *filename_body = NULL;
 	struct stat sbuf;
-	int position = 0;
 	int start_position = 0;
 	int text_handler = 0;
 	int temp_fd = -1;
@@ -273,12 +271,11 @@ static int layout_handler (request_rec * r) {
 	if (r->main) {
 		return DECLINED;
 	}
-
 	/* Since I keep finding that people want me to help debug, yet 
 		have their server only listing Apache. This of course came
 		from the nice folks at PHP over a night of beers at a conference.
 	*/
-	ap_table_setn(r->headers_out, "X-Powered-BY", "ModLayout/"VERSION);
+	ap_table_setn(r->headers_out, "X-Powered-By", "ModLayout/"VERSION);
 
 	cfg = ap_get_module_config(r->per_dir_config, &layout_module);
 	info = ap_get_module_config(r->request_config, &layout_module);
@@ -290,22 +287,22 @@ static int layout_handler (request_rec * r) {
 	pid = r->connection->child_num;
 
 /********************************************************************/
-	if(isOn(cfg->async_post) && info->length) {
+	if (isOn(cfg->async_post) && info->length) {
 		filename_post = ap_psprintf(r->pool, "%s/.mod_layout.post.%d",cfg->async_cache, pid);
 		/* Now lets suck up some content */
-		if((status = read_content(r, filename_post, info->length)) != OK) {
+		if ((status = read_content(r, filename_post, info->length)) != OK) {
 			ap_log_rerror(APLOG_MARK, APLOG_NOERRNO|APLOG_ERR, r,
 				"mod_layout got an error while doing async post : %d", status);
 		}
 		/* Let us be paranoid about content length */
 		/* Should try to cut this out (one less stat) */
-		if(stat(filename_post, &sbuf)) {
+		if (stat(filename_post, &sbuf)) {
 			/* This would be very bad */
 			status = HTTP_INTERNAL_SERVER_ERROR;
 			ap_log_rerror(APLOG_MARK, APLOG_NOERRNO | APLOG_ERR, r, 
 					"fstat blew chunks in mod_layout for file: %s", filename_post);
 		}
-		if((status = get_fd_in(r, filename_post)) != OK) {
+		if ((status = get_fd_in(r, filename_post)) != OK) {
 			ap_log_rerror(APLOG_MARK, APLOG_NOERRNO|APLOG_ERR, r,
 					"mod_layout couldn't open a file descriptor for the post");
 			return status;
@@ -315,10 +312,9 @@ static int layout_handler (request_rec * r) {
 
 /********************************************************************/
 /* First, we get the original document */
-	if(isOn(info->origin)) {
+	if (isOn(info->origin)) {
 		ap_rflush(r);
-		if(isOn(cfg->layout_html_handler) && check_type(info)) {
-WATCHPOINT
+		if (isOn(cfg->layout_html_handler) && check_type(info)) {
 			filename_body = r->filename;
 			text_handler++;
 			if ((temp_fd = ap_popenf(r->pool, filename_body,O_RDONLY,S_IRWXU)) < 0) {
@@ -329,30 +325,33 @@ WATCHPOINT
 			}
 
 		} else {
-WATCHPOINT
 			buff = ap_bcreate(r->pool, B_RDWR);
 			obuff = r->connection->client;
-			buff->fd_in = r->connection->client->fd_in;
+			buff->fd_in = dup(r->connection->client->fd_in);
 			buff->incnt = r->connection->client->incnt;
 			buff->inptr = r->connection->client->inptr;
 
 			filename_body = ap_psprintf(r->pool, "%s/.mod_layout.body.%d",cfg->async_cache, getpid());
 
-			if((status = get_fd_out(r, filename_body, buff)) != OK) {
+			if ((status = get_fd_out(r, filename_body, buff)) != OK) {
 				return status;
 			}
-			temp_fd = buff->fd;
 			r->connection->client = buff;
 
 			status = layout_origin(r, cfg, info);
 
-			lseek(temp_fd, 0, SEEK_SET);
-			buff->fd = -1;
-			buff->fd_in = -1;
+			ap_bflush(buff);
+			temp_fd = dup(buff->fd);
+			//temp_fd = buff->fd;
+			fsync(temp_fd);
+			//buff->fd = -1;
+			//buff->fd_in = -1;
 			ap_bclose(buff);
+			lseek(temp_fd, 0, SEEK_SET);
 			r->connection->client = obuff;
-			if(status != OK) {
+			if (status != OK) {
 				layout_send_file(r, filename_body);
+				close(temp_fd);
 				return status;
 			}
 		}
@@ -361,13 +360,15 @@ WATCHPOINT
 		map_data = ap_pcalloc (r->pool, sizeof (mmap_data));
 
 	/* Dump the HTTP crap */
-		if(text_handler) {
+		/*
+		if (text_handler) {
 			map_data->size = r->finfo.st_size;
 		} else {
+		*/
 			(void)fstat(temp_fd, &sbuf);
 			map_data->size = sbuf.st_size;
-		}
-		body = (char *)mmap(NULL, map_data->size, PROT_READ, MAP_PRIVATE, temp_fd, 0);
+	//	}
+		body = mmap(NULL, map_data->size, PROT_READ, MAP_PRIVATE, temp_fd, 0);
 		map_data->file = body;
 		ap_register_cleanup(r->pool, map_data, cleanup_mmap, ap_null_cleanup);
 
@@ -381,14 +382,16 @@ WATCHPOINT
 		/* No reason to not enable this for all types of connections */
 		/* We should check this before we scan headers */
 
-		if(is_ignored(r, cfg, info, body)) {
-			ap_send_mmap(map_data->file, r, 0, map_data->size);
+		if (is_ignored(r, cfg, info, body)) {
+			//ap_send_mmap(map_data->file, r, 0, map_data->size);
+			ap_rputs(body, r);
+			close(temp_fd);
 			return DONE;
 		}
 
-		if(!text_handler) {
+		if (!text_handler) {
 			start_position = find_headers(r, body);
-			if(start_position < 0) { 
+			if (start_position < 0) { 
 				start_position = 0;
 			}
 		}
@@ -398,109 +401,82 @@ WATCHPOINT
 	printf("Start position is %d\n", start_position);
 #endif
 
-	if(isOn(cfg->notes))
+	if (isOn(cfg->notes))
 		update_info(r->notes, info);
 
 	if (info->http_header){
 		if ((status = call_container(r, cfg->http_header, cfg, info, 0)) != OK){
 			ap_log_rerror(APLOG_MARK, APLOG_NOERRNO | APLOG_ERR, r, "The following error occured while processing the Layout HTTP Header : %d", status); 
 		}
+		ap_rflush(r);
 	} else {
 		print_layout_headers(r, cfg);
 	}
 	ap_rflush(r);
-	if(isOn(cfg->notes))
+
+	if (isOn(cfg->notes))
 		update_info(r->notes, info);
 
 
 /********************************************************************/
 /* Welcome to the problem, lets patch this all back up */
-	if(isOn(info->merge) && info->origin) {
-
+	if (isOn(info->merge) && info->origin) {
 #ifdef DEBUG
 		printf("Merge Header:%d: Footer:%d: HTTPHeader:%d:\n", info->header, info->footer, info->http_header);
 #endif
 
 /* Look into using ap_send_mmap() -Brian */
-		x = start_position;
-		if(info->header)  {
-			if((position = string_search(r, body, cfg->begin_tag, start_position, (isOn(cfg->replace_tags) ? 1 : 0))) != -1) {
-				for(x = start_position; x < position; x++) {
-					ap_rputc(body[x], r);
-				}
-				layout_header(r, cfg, info);
-				if(isOn(cfg->notes))
-					update_info(r->notes, info);
-				if(isOn(cfg->replace_tags)) {
-					position = string_search(r, body, cfg->begin_tag, position, 0);
-				}
-				x = position;
-			} else {
-				layout_header(r, cfg, info);
-				if(isOn(cfg->notes))
-					update_info(r->notes, info);
-				x = start_position;
+		if (isOn(info->origin))  {
+			/* This code right here could be replaced and this would all be faster */
+			/* We have to make two passes through to make sure the headers always happen */
+			if (info->header && (string_search(r, body, cfg->begin_tag, start_position, 0) == -1)) {	
+				layout_kind(r, cfg, info, HEADER);
 			}
+			parser_put(r, cfg, info,  body, start_position);
+			if (info->footer && (string_search(r, body, cfg->end_tag, start_position, 0) == -1)) {	
+				layout_kind(r, cfg, info, FOOTER);
+			}
+		} else {
+			layout_kind(r, cfg, info, HEADER);
+			if (isOn(cfg->notes))
+				update_info(r->notes, info);
+			layout_kind(r, cfg, info, LAYOUT);
+			if (isOn(cfg->notes))
+				update_info(r->notes, info);
+			layout_kind(r, cfg, info, FOOTER);
 		}
 
-		if(info->footer && isOn(info->origin))  {
-			if((position = string_search(r, body, cfg->end_tag, x, 1)) != -1) {
-				for(; x < position; x++) {
-					ap_rputc(body[x], r);
-				}
-				layout_footer(r, cfg, info);
-				if(isOn(cfg->replace_tags)) {
-					x = string_search(r, body, cfg->end_tag, x, 0);
-				}
-				for(; x < map_data->size; x++) {
-					ap_rputc(body[x], r);
-				}
-			} else {
-				for(; x < map_data->size; x++) {
-					ap_rputc(body[x], r);
-				}
-				layout_footer(r, cfg, info);
-				if(isOn(cfg->notes))
-					update_info(r->notes, info);
-			}
-		} else if(isOn(info->origin)) {
-				for(; x < map_data->size; x++) {
-					ap_rputc(body[x], r);
-				}
-		} else if(isOff(info->origin) && info->footer) {
-				layout_footer(r, cfg, info);
-				if(isOn(cfg->notes))
-					update_info(r->notes, info);
-		}
 	} else {
 /********************************************************************/
 /* This is the original stream method */
 		/* Now we do headers if they exist */
 		if (info->header) {
-			layout_header(r, cfg, info);
+			layout_kind(r, cfg, info, HEADER);
 		}
 		ap_rflush(r);
-		if(isOn(cfg->notes))
+		if (isOn(cfg->notes))
 			update_info(r->notes, info);
 
 	/********************************************************************/
 		/* Now we handle the orignal request.  */
 
 		if (isOn(info->origin)) {
-			ap_send_mmap(map_data->file, r, start_position, map_data->size);
+			//ap_send_mmap(map_data->file, r, start_position, map_data->size);
+			ap_rputs(body + start_position, r);
 		}
 		ap_rflush(r);
 
-		if(isOn(cfg->notes))
+		if (isOn(cfg->notes))
 			update_info(r->notes, info);
 
 	/********************************************************************/
 	/* Now lets do the footer */
 		if (info->footer) {
-			layout_footer(r, cfg, info);
+			layout_kind(r, cfg, info, FOOTER);
 		}
 		ap_rflush(r);
 	}
+	close(temp_fd);
 
 	return DONE;
 }
@@ -512,7 +488,8 @@ static int layout_fixup(request_rec *r) {
 	int length; 
 	layout_request *info = NULL;
 
-	if (isOff(cfg->footer_enabled) && isOff(cfg->header_enabled) && isOff(cfg->http_header_enabled)) {
+	/* Bug here, what about random layouts? */
+	if (isOff(cfg->footer_enabled) && isOff(cfg->header_enabled) && isOff(cfg->http_header_enabled) && isOff(cfg->pattern_enabled)) {
 		return DECLINED;
 	}
 	if (r->main) {
@@ -525,8 +502,8 @@ static int layout_fixup(request_rec *r) {
 
 /* Now, lets fix it if some goof has called a URL that 
    should have had a / in it */
-	if(ap_is_directory(r->filename)) {
-		if(r->uri[0] == '\0' || r->uri[strlen(r->uri) - 1] != '/') {
+	if (ap_is_directory(r->filename)) {
+		if (r->uri[0] == '\0' || r->uri[strlen(r->uri) - 1] != '/') {
 /* Now at this point we know things are not going to
    go over well, so lets just let it all die in some module
    designed to take care of this sort of thing */
@@ -539,7 +516,7 @@ static int layout_fixup(request_rec *r) {
 		 has changed about the way that CGI's are done. Not sure what
 		 it is, but this is now needed */
 	/* First, we check to see if this is SSI, mod_perl or cgi */
-	if(r->handler) {
+	if (r->handler) {
 		type = ap_pstrdup(r->pool, r->handler);
 	} else {
 		type = ap_pstrdup(r->pool, r->content_type);
@@ -557,7 +534,7 @@ static int layout_fixup(request_rec *r) {
 		from perfect and is still pretty much a shot in the dark.
 		More research needed.
 		*/
-		if(r->uri[length - 1] == '/' ) {
+		if (r->uri[length - 1] == '/' ) {
 			type = "text/html";
 		} else {
 			subr = (request_rec *) ap_sub_req_lookup_file(r->uri, r);
@@ -570,7 +547,7 @@ static int layout_fixup(request_rec *r) {
 
 	/* Now lets look at the ignore logic */
 	/* This is where we finally decide what is going to happen */
-	if(cfg->uris_ignore) {
+	if (cfg->uris_ignore) {
 		if (table_find(cfg->uris_ignore, r->uri))
 			return DECLINED;
 	}
@@ -611,7 +588,7 @@ static const char *add_http_header(cmd_parms * cmd, void *mconfig, char *uri) {
 	return NULL;
 }
 
-static const char *add_layout(cmd_parms * cmd, void *mconfig, char *param) {
+static const char *add_layout_pattern(cmd_parms * cmd, void *mconfig, char *pattern, char *layout, char *mode) {
 	layout_conf *cfg = (layout_conf *) mconfig;
 	struct stat sbuf;
 	const char *temp;
@@ -620,39 +597,69 @@ static const char *add_layout(cmd_parms * cmd, void *mconfig, char *param) {
 
 	directive = ap_pcalloc (cmd->pool, sizeof (layout_string));
 
-	if(ap_ind(param, ' ') != -1) {
+	if (ap_ind(layout, ' ') != -1) {
 		directive->comment = ap_pstrdup (cmd->pool, "Static Content");
-		directive->string = ap_pstrdup (cmd->pool, param);
+		directive->string = ap_pstrdup (cmd->pool, layout);
 		directive->type = 1;
-	} else if(stat(param, &sbuf) == 0){
-		unless((temp = add_file(cmd, cfg, param))) {
+	} else if (stat(layout, &sbuf) == 0){
+		unless((temp = add_file(cmd, cfg, layout))) {
 			return NULL;
 		};
-		directive->comment = ap_pstrdup (cmd->pool, param);
+		directive->comment = ap_pstrdup (cmd->pool, layout);
 		directive->string = ap_pstrdup (cmd->pool, temp);
 		directive->type = 1;
 	} else {
-		directive->comment = ap_pstrdup (cmd->pool, param);
-		directive->string = ap_pstrdup (cmd->pool, param);
+		directive->comment = ap_pstrdup (cmd->pool, layout);
+		directive->string = ap_pstrdup (cmd->pool, layout);
 		directive->type = 0;
-		if(cfg->cache_needed == UNSET)
+		if (cfg->cache_needed == UNSET)
 			cfg->cache_needed = ON;
 	}
+	directive->pattern = ap_pstrdup (cmd->pool, pattern);
+
+	unless(cfg->layouts) {
+		cfg->layouts = ap_make_array (cmd->pool, 1, sizeof (layout_string *));
+	}
 
 	unless(strcasecmp(cmd->cmd->name, "LayoutHeader")) {
-		unless(cfg->header) {
-			cfg->header = ap_make_array (cmd->pool, 1, sizeof (layout_string *));
-		}
-		*(layout_string **) ap_push_array (cfg->header) = (layout_string *) directive;
 		cfg->header_enabled = ON;
+		directive->kind = HEADER;
 	} else unless(strcasecmp(cmd->cmd->name, "LayoutFooter")) {
-		unless(cfg->footer) {
-			cfg->footer = ap_make_array (cmd->pool, 1, sizeof (layout_string *));
-		}
-		*(layout_string **) ap_push_array (cfg->footer) = (layout_string *) directive;
 		cfg->footer_enabled = ON;
+		directive->kind = FOOTER;
+	} else {
+		cfg->pattern_enabled = ON;
+		directive->kind = LAYOUT;
 	}
 
+	if (mode) {
+		unless(strcasecmp(mode, "append")) {
+			directive->append = APPEND;
+		} else unless(strcasecmp(mode, "prepend")) {
+			directive->append = PREPEND;
+		} else unless(strcasecmp(mode, "replace")) {
+			directive->append = REPLACE;
+		} else {
+			ap_log_error (APLOG_MARK, APLOG_ERR, cmd->server,
+				"The type %s is not valid for %s ",mode, cmd->cmd->name);
+			directive->append = REPLACE;
+		}
+	} else {
+		directive->append = REPLACE;
+	}
+
+	*(layout_string **) ap_push_array (cfg->layouts) = (layout_string *) directive;
+
+	return NULL;
+}
+
+static const char *add_layout(cmd_parms * cmd, void *mconfig, char *layout) {
+	layout_conf *cfg = (layout_conf *) mconfig;
+	unless(strcasecmp(cmd->cmd->name, "LayoutHeader")) {
+			(void)add_layout_pattern(cmd, mconfig, (char *)cfg->begin_tag, layout, "append");
+		} else unless(strcasecmp(cmd->cmd->name, "LayoutFooter")) {
+			(void)add_layout_pattern(cmd, mconfig, (char *)cfg->end_tag, layout, "prepend");
+	}
 	return NULL;
 }
 
@@ -704,32 +711,32 @@ static const char *add_type(cmd_parms * cmd, void *mconfig, char *type) {
 	return NULL;
 }
 
-static const char *merge_ignore_add(cmd_parms * cmd, void *mconfig, char *type) {
+static const char *tag_ignore_add(cmd_parms * cmd, void *mconfig, char *type) {
 	layout_conf *cfg = (layout_conf *) mconfig;
-	unless(cfg->merge_ignore)
-		cfg->merge_ignore = ap_make_table(cmd->pool, 1);
+	unless(cfg->tag_ignore)
+		cfg->tag_ignore = ap_make_table(cmd->pool, 1);
 
-	ap_table_setn(cfg->merge_ignore, type, "1");
+	ap_table_setn(cfg->tag_ignore, type, "1");
 
 	return NULL;
 }
 
-static const char *merge_ignore_footer_add(cmd_parms * cmd, void *mconfig, char *type) {
+static const char *tag_ignore_footer_add(cmd_parms * cmd, void *mconfig, char *type) {
 	layout_conf *cfg = (layout_conf *) mconfig;
-	unless(cfg->merge_ignore_footer)
-		cfg->merge_ignore_footer = ap_make_table(cmd->pool, 1);
+	unless(cfg->tag_ignore_footer)
+		cfg->tag_ignore_footer = ap_make_table(cmd->pool, 1);
 
-	ap_table_setn(cfg->merge_ignore_footer, type, "1");
+	ap_table_setn(cfg->tag_ignore_footer, type, "1");
 
 	return NULL;
 }
 
-static const char *merge_ignore_header_add(cmd_parms * cmd, void *mconfig, char *type) {
+static const char *tag_ignore_header_add(cmd_parms * cmd, void *mconfig, char *type) {
 	layout_conf *cfg = (layout_conf *) mconfig;
-	unless(cfg->merge_ignore_header)
-		cfg->merge_ignore_header = ap_make_table(cmd->pool, 1);
+	unless(cfg->tag_ignore_header)
+		cfg->tag_ignore_header = ap_make_table(cmd->pool, 1);
 
-	ap_table_setn(cfg->merge_ignore_header, type, "1");
+	ap_table_setn(cfg->tag_ignore_header, type, "1");
 
 	return NULL;
 }
@@ -759,7 +766,7 @@ static const char *header_off(cmd_parms * cmd, void *mconfig) {
 
 static const char *http_header_off(cmd_parms * cmd, void *mconfig) {
 	layout_conf *cfg = (layout_conf *) mconfig;
-	cfg->header = NULL;
+	cfg->http_header = NULL;
 	cfg->http_header_enabled = OFF;
 
 	return NULL;
@@ -785,6 +792,8 @@ static const char *remove_default_types(cmd_parms * cmd, void *mconfig, int flag
 	ap_table_setn(cfg->types, "text/plain", "0");
 	ap_table_setn(cfg->types, "perl-script", "0");
 	ap_table_setn(cfg->types, "application/x-httpd-php", "0");
+	ap_table_setn(cfg->types, "application/x-httpd-php3", "0");
+	ap_table_setn(cfg->types, "application/x-httpd-php-source", "0");
 	ap_table_setn(cfg->types, "cgi-script", "0");
 	ap_table_setn(cfg->types, "application/x-httpd-cgi", "0");
 
@@ -792,11 +801,12 @@ static const char *remove_default_types(cmd_parms * cmd, void *mconfig, int flag
 }
 
 static const command_rec layout_cmds[] = {
+	{"Layout", add_layout_pattern, NULL, OR_ALL, TAKE23, Layout},
 	{"LayoutHeader", add_layout, NULL, OR_ALL, TAKE1, LayoutHeader},
-	{"LayoutAppend", append_layouts, NULL, OR_ALL, FLAG, LayoutAppend},
-	{"LayoutHeaderAppend", ap_set_flag_slot, (void *) XtOffsetOf(layout_conf, append_header), OR_ALL, FLAG, LayoutHeaderAppend},
-	{"LayoutFooterAppend", ap_set_flag_slot, (void *) XtOffsetOf(layout_conf, append_footer), OR_ALL, FLAG, LayoutFooterAppend},
 	{"LayoutFooter", add_layout, NULL, OR_ALL, TAKE1, LayoutFooter},
+	{"LayoutAppend", append_layouts, NULL, OR_ALL, FLAG, LayoutAppend},
+	{"LayoutAppendHeader", ap_set_flag_slot, (void *) XtOffsetOf(layout_conf, append_header), OR_ALL, FLAG, LayoutHeaderAppend},
+	{"LayoutAppendFooter", ap_set_flag_slot, (void *) XtOffsetOf(layout_conf, append_header), OR_ALL, FLAG, LayoutFooterAppend},
 	{"LayoutHandler", add_type, NULL, OR_ALL, TAKE1, LayoutHandler},
 	{"LayoutIgnoreURI", ignore_uri, NULL, OR_ALL, TAKE1, LayoutIgnoreURI},
 	{"LayoutIgnoreHeaderURI", ignore_header_uri, NULL, OR_ALL, TAKE1,
@@ -815,9 +825,9 @@ static const command_rec layout_cmds[] = {
 	{"LayoutTimeFormat", ap_set_string_slot, (void *) XtOffsetOf(layout_conf, time_format),
 	 OR_ALL, TAKE1, LayoutTimeFormat},
 	{"LayoutHTTPHeader", add_http_header, NULL, OR_ALL, TAKE1, LayoutHTTPHeader},
-	{"LayoutIgnoreTag", merge_ignore_add, NULL, OR_ALL, TAKE1, LayoutIgnoreTag},
-	{"LayoutIgnoreTagFooter", merge_ignore_footer_add, NULL, OR_ALL, TAKE1, LayoutIgnoreTagFooter},
-	{"LayoutIgnoreTagHeader", merge_ignore_header_add, NULL, OR_ALL, TAKE1, LayoutIgnoreTagHeader},
+	{"LayoutIgnoreTag", tag_ignore_add, NULL, OR_ALL, TAKE1, LayoutIgnoreTag},
+	{"LayoutIgnoreTagFooter", tag_ignore_footer_add, NULL, OR_ALL, TAKE1, LayoutIgnoreTagFooter},
+	{"LayoutIgnoreTagHeader", tag_ignore_header_add, NULL, OR_ALL, TAKE1, LayoutIgnoreTagHeader},
 	{"LayoutFooterOff", footer_off, NULL , OR_ALL, NO_ARGS, LayoutFooterOff},
 	{"LayoutHeaderOff", header_off, NULL, OR_ALL, NO_ARGS, LayoutHeaderOff},
 	{"LayoutHTTPHeaderOff", http_header_off, NULL, OR_ALL, NO_ARGS, LayoutHTTPHeaderOff},
