@@ -1,6 +1,6 @@
 /* 
 **  utility.c -- Apache layout module
-**  $Revision: 1.51 $
+**  $Revision: 1.57 $
 */
 
 #include "mod_layout.h"
@@ -10,8 +10,33 @@ LAYOUT_EXPORT(void) cleanup_mmap(void *data) {
 	munmap(file->file, file->size);
 }
 
+LAYOUT_EXPORT(array_header *) layout_array_push_kind(pool *p,array_header *origin, array_header *new, int kind) {
+	array_header *returnable = NULL;
+	int x = 0;
+	layout_string **layouts;
+
+	layouts = (layout_string **) new->elts;
+
+	if (!origin && !new) {
+		return NULL;
+	} else if (!origin) {
+		return new;
+	} else if (!new) {
+		return origin;
+	}
+
+	/* Size of new plus some padding */
+	returnable = ap_make_array (p, new->nelts + 2, sizeof (layout_string *));
+	for (x = 0; x < new->nelts; x++) {
+		if (layouts[x]->kind == kind) {
+			*(layout_string **) ap_push_array (returnable) = (layout_string *) layouts[x];
+		}
+	}
+	ap_array_cat(returnable, new);
+	return returnable;
+}
+
 LAYOUT_EXPORT(int) check_type(layout_request *request) {
-WATCHPOINT
 #ifdef DEBUG
 	printf("check_type: %s \n", request->type);
 #endif
@@ -46,6 +71,10 @@ LAYOUT_EXPORT(layout_request *) create_layout_request(request_rec *r, layout_con
 
 	info->type = type;
 
+#ifdef DEBUG
+	//printf("IGNORES HTTP %d Header %d Footer %d\n", cfg->uris_ignore_http_header, cfg->uris_ignore_header, cfg->footer_enabled);
+#endif
+
 	if(isOn(cfg->header_enabled)) {
 		info->header = ON;
 		if(cfg->uris_ignore_header) {
@@ -94,7 +123,7 @@ LAYOUT_EXPORT(int) table_search(request_rec *r, const table * t, const char *str
 	return 1;
 }
 
-void table_cat(table *src, table *dest, char *string) {
+LAYOUT_EXPORT(void) table_cat(table *src, table *dest, char *string) {
 	array_header *hdrs_arr;
 	table_entry *elts;
 	int x;
@@ -181,6 +210,87 @@ LAYOUT_EXPORT(int) string_search(request_rec *r, char *string, const char *delim
 	return -1;
 }
 
+LAYOUT_EXPORT(int) parser_put(request_rec *r, layout_conf *cfg, layout_request *info, char *string, int init_pos) {
+	char *temp = NULL;
+	char *sub_temp = NULL;
+	char *lower = NULL;
+	char *substring = NULL;
+	int length = 0;
+	int end = 0;
+	int x = 0;
+	int j = 0;
+	int match = 0;
+	int complete_position = 0;
+	int run = 0;
+	layout_string **layouts;
+
+	layouts = (layout_string **) cfg->layouts->elts;
+
+	if(string == NULL) 
+		return -1;
+
+	length = strlen(string);
+
+	temp = string + init_pos;
+	complete_position = init_pos;
+
+	x = init_pos;
+	while ( x < length ) {
+		if( string[x] == '<') {
+			sub_temp = string + x;
+			if((end = ap_ind(sub_temp, '>')) != -1) { 
+				substring = ap_pstrndup(r->pool, sub_temp , end + 1);
+				lower = ap_pstrdup(r->pool, substring);
+				ap_str_tolower(lower);
+				for (j = 0; j < cfg->layouts->nelts; j++) {
+					/* The next few lines should be reduced */
+					run = 1;
+					if (layouts[j]->kind == HEADER && !info->header) {
+						run = 0;
+					}
+					if (layouts[j]->kind == FOOTER && !info->footer) {
+						run = 0;
+					}
+					if (run) {
+						unless(ap_fnmatch(layouts[j]->pattern, lower, FNM_CASE_BLIND)) {
+							if(layouts[j]->append == APPEND) {
+								ap_rputs(substring,r);
+								layout_print(r, cfg, info, j);
+								if(isOn(cfg->notes))
+									update_info(r->notes, info);
+							} else if(layouts[j]->append == PREPEND) {
+								layout_print(r, cfg, info, j);
+								if(isOn(cfg->notes))
+									update_info(r->notes, info);
+								ap_rputs(substring, r);
+							} else { /* We replace! */
+								layout_print(r, cfg, info, j);
+								if(isOn(cfg->notes))
+									update_info(r->notes, info);
+							}
+							match++;
+						}
+					}
+				}
+				unless (match)  {
+					ap_rputs(substring, r);
+				}
+				match = 0;
+				x += strlen(substring);
+			} else {
+				/* If we never match just write out the substring and keep going */
+				ap_rputc(string[x], r);
+				x++;
+			}
+		} else {
+			ap_rputc(string[x], r);
+			x++;
+		}
+	}
+
+	return -1;
+}
+
 LAYOUT_EXPORT(int) find_headers(request_rec *r, char *body) {
 	int position = 0;
 	char *temp = NULL;
@@ -252,7 +362,7 @@ LAYOUT_EXPORT(int) layout_send_file(request_rec *r, char *filename) {
 
 LAYOUT_EXPORT(int) get_fd_out(request_rec *r, char *filename, BUFF *buff) {
 	int status = OK;
-	int temp_fd = 0;
+	int temp_fd = -1;
 
 	if ((temp_fd = ap_popenf(r->pool, filename,O_RDWR|O_CREAT|O_TRUNC,S_IRWXU)) == -1) {
 		ap_log_rerror(APLOG_MARK, APLOG_NOERRNO|APLOG_ERR, r,
@@ -374,20 +484,20 @@ LAYOUT_EXPORT(void) update_info(const table *t, layout_request *info) {
 }
 
 LAYOUT_EXPORT(int) is_ignored(request_rec *r, layout_conf *cfg, layout_request *info, char *body) {
-	if(cfg->merge_ignore) {
-		if(table_search(r, cfg->merge_ignore, body)){
+	if(cfg->tag_ignore) {
+		if(table_search(r, cfg->tag_ignore, body)){
 			info->header = OFF;
 			info->footer = OFF;
 			return 1;
 		}
 	}
-	if(cfg->merge_ignore_footer) {
-		if(table_search(r, cfg->merge_ignore_footer, body)){
+	if(cfg->tag_ignore_footer) {
+		if(table_search(r, cfg->tag_ignore_footer, body)){
 			info->footer = OFF;
 		}
 	}
-	if(cfg->merge_ignore_header) {
-		if(table_search(r, cfg->merge_ignore_header, body)){
+	if(cfg->tag_ignore_header) {
+		if(table_search(r, cfg->tag_ignore_header, body)){
 			info->header = OFF;
 		}
 	}
@@ -487,7 +597,7 @@ LAYOUT_EXPORT(void) print_layout_headers(request_rec *r, layout_conf *cfg) {
 
 
 
-LAYOUT_EXPORT(char *)add_file(cmd_parms * cmd, layout_conf *cfg, char *file) {
+LAYOUT_EXPORT(char *) add_file(cmd_parms * cmd, layout_conf *cfg, const char *file) {
 	FILE *file_ptr;
 	char buf[HUGE_STRING_LEN];
 	char *content = NULL;
@@ -510,9 +620,22 @@ LAYOUT_EXPORT(char *)add_file(cmd_parms * cmd, layout_conf *cfg, char *file) {
 	return content;
 }
 
-LAYOUT_EXPORT(void)reset_fd(request_rec *r, int length) {
+LAYOUT_EXPORT(void) reset_fd(request_rec *r, int length) {
 	r->remaining = length;
 	r->read_length = 0;
 	r->read_chunked = 0;
 	lseek(r->connection->client->fd_in, 0, SEEK_SET);
 }
+
+LAYOUT_EXPORT(void) layout_kind(request_rec *r, layout_conf *cfg, layout_request *info, int kind) {
+	int x = 0;
+	layout_string **layouts;
+
+	layouts = (layout_string **) cfg->layouts->elts;
+
+	for(x = 0; x < cfg->layouts->nelts; x++) {
+		if (layouts[x]->kind == kind) {
+			layout_print(r, cfg, info, x);
+		}
+	}
+}
